import { BlobClientResponse, ClientMode, JsonClientResponse, TextClientResponse, VoidClientResponse } from '../client';
import { FetchClient } from '../ws';
import { BriefcaseTemplate } from './BriefcaseTemplate';
import { PersistenceTools } from './PersistenceTools';
export class DialogProxy {
    constructor(clientMode = ClientMode.ONLINE) {
        /* Last operation happened at this time */
        this._lastActivity = new Date();
        this._clientMode = clientMode;
        this._fetchClient = new FetchClient();
        const PersistentClientVars = JSON.parse(window.localStorage.getItem('PersistentClientVars'));
        if (PersistentClientVars) {
            this._tenantId = PersistentClientVars.tenantId;
            this._userId = PersistentClientVars.userId;
            this._sessionId = PersistentClientVars.sessionId;
            this._clientMode = PersistentClientVars.clientMode === 'OFFLINE' ? ClientMode.OFFLINE : ClientMode.ONLINE;
        }
    }
    get lastActivity() {
        return this._lastActivity;
    }
    writePersistentClientVars() {
        const PersistentClientVars = {
            tenantId: this._tenantId,
            userId: this._userId,
            sessionId: this._sessionId,
            clientMode: this._clientMode === ClientMode.OFFLINE ? 'OFFLINE' : 'ONLINE'
        };
        window.localStorage.setItem('PersistentClientVars', JSON.stringify(PersistentClientVars));
    }
    changeClientMode(clientMode) {
        this._clientMode = clientMode;
        this.writePersistentClientVars();
    }
    changeUserId(userId) {
        this._userId = userId;
        this.writePersistentClientVars();
    }
    changeSessionId(sessionId) {
        this._sessionId = sessionId;
        this.writePersistentClientVars();
    }
    changeTenantId(tenantId) {
        this._tenantId = tenantId;
        this.writePersistentClientVars();
    }
    createDialogMessageModel(message) {
        return { type: 'hxgn.api.dialog.DialogMessage', message };
    }
    createOfflineSessionNotFoundModel() {
        return this.createDialogMessageModel('Offline session not found');
    }
    createSessionIdModel(sessionId) {
        return { type: 'hxgn.api.dialog.SessionId', sessionId };
    }
    deconstructGetDialogPath(resourcePathElems) {
        return {
            tenantId: resourcePathElems[1],
            sessionId: resourcePathElems[3],
            dialogId: resourcePathElems[5]
        };
    }
    deconstructGetRecordPath(resourcePathElems) {
        return {
            tenantId: resourcePathElems[1],
            sessionId: resourcePathElems[3],
            dialogId: resourcePathElems[5]
        };
    }
    deconstructGetRedirectionPath(resourcePathElems) {
        return {
            tenantId: resourcePathElems[1],
            sessionId: resourcePathElems[3],
            redirectionId: resourcePathElems[5]
        };
    }
    deconstructPostMenuActionPath(resourcePathElems) {
        return {
            tenantId: resourcePathElems[1],
            sessionId: resourcePathElems[3],
            dialogId: resourcePathElems[5],
            actionId: resourcePathElems[7]
        };
    }
    deconstructPostRecordsPath(resourcePathElems) {
        return {
            tenantId: resourcePathElems[1],
            sessionId: resourcePathElems[3],
            dialogId: resourcePathElems[5]
        };
    }
    deconstructPostWorkbenchActionPath(resourcePathElems) {
        return {
            tenantId: resourcePathElems[1],
            sessionId: resourcePathElems[3],
            workbenchId: resourcePathElems[5],
            actionId: resourcePathElems[7]
        };
    }
    // @TODO
    openStream(baseUrl, resourcePath) {
        return Promise.resolve(null);
    }
    deleteJson(baseUrl, resourcePath) {
        const resourcePathElems = resourcePath.split('/');
        if (this._clientMode === ClientMode.OFFLINE) {
            if (PersistenceTools.isDeleteSession(resourcePathElems)) {
                return this.deleteSession(baseUrl, resourcePath);
            }
            else {
                return Promise.resolve(new JsonClientResponse(this.createDialogMessageModel(`DELETE action is not valid while offline: ${resourcePath}`), 400));
            }
        }
        return this._fetchClient.deleteJson(baseUrl, resourcePath);
    }
    deleteSession(baseUrl, resourcePath) {
        return new Promise((resolve, reject) => {
            const session = PersistenceTools.readSessionState(this._tenantId, this._userId);
            if (!session) {
                resolve(new JsonClientResponse(this.createOfflineSessionNotFoundModel(), 404));
            }
            else {
                const sessionIdModel = this.createSessionIdModel(session.id);
                resolve(new JsonClientResponse(sessionIdModel, 200));
            }
        });
    }
    getBlob(baseUrl, resourcePath) {
        if (this._clientMode === ClientMode.OFFLINE) {
            return Promise.resolve(new BlobClientResponse(null, 400));
        }
        return this._fetchClient.getBlob(baseUrl, resourcePath);
    }
    getDialog(baseUrl, resourcePath, resourcePathElems, queryParams) {
        const pathFields = this.deconstructGetDialogPath(resourcePathElems);
        if (pathFields.dialogId === DialogProxy.BRIEFCASE_DIALOG_ID) {
            let dialog = PersistenceTools.readDialogState(this._tenantId, this._userId, pathFields.dialogId);
            if (!dialog) {
                dialog = BriefcaseTemplate.BRIEFCASE_DIALOG;
                dialog.sessionId = this._sessionId;
                PersistenceTools.writeDialogState(this._tenantId, this._userId, dialog);
                PersistenceTools.writeAllDialogParentState(this._tenantId, this._userId, dialog);
            }
            return Promise.resolve(new JsonClientResponse(dialog, 200));
        }
        if (this._clientMode === ClientMode.OFFLINE) {
            const dialog = PersistenceTools.readDialogState(this._tenantId, this._userId, pathFields.dialogId);
            if (!dialog) {
                return Promise.resolve(new JsonClientResponse(this.createDialogMessageModel(`Offline dialog not found: ${pathFields.dialogId}`), 404));
            }
            else {
                return Promise.resolve(new JsonClientResponse(dialog, 200));
            }
        }
        const response = this._fetchClient.getJson(baseUrl, resourcePath, queryParams);
        return response.then(jcr => {
            if (jcr.statusCode === 200) {
                let dialog = jcr.value;
                if (this.isWorkPackagesDialog(dialog)) {
                    dialog = this.patchWorkPackagesDialog(dialog);
                }
                // If a redirection object has been persisted, then also persist the dialog object.
                const redirection = PersistenceTools.readRedirectionState(this._tenantId, this._userId, dialog.id);
                if (redirection) {
                    PersistenceTools.writeDialogState(this._tenantId, this._userId, dialog);
                    PersistenceTools.writeAllDialogParentState(this._tenantId, this._userId, dialog);
                }
                return new JsonClientResponse(dialog, 200);
            }
            return response;
        });
    }
    getJson(baseUrl, resourcePath, queryParams) {
        const resourcePathElems = resourcePath.split('/');
        if (PersistenceTools.isGetDialog(resourcePathElems)) {
            return this.getDialog(baseUrl, resourcePath, resourcePathElems, queryParams);
        }
        else if (PersistenceTools.isGetRedirection(resourcePathElems)) {
            return this.getRedirection(baseUrl, resourcePath, resourcePathElems, queryParams);
        }
        else if (PersistenceTools.isGetSession(resourcePathElems)) {
            return this.getSession(baseUrl, resourcePath, resourcePathElems, queryParams);
        }
        else if (PersistenceTools.isGetRecord(resourcePathElems)) {
            return this.getRecord(baseUrl, resourcePath, resourcePathElems, queryParams);
        }
        if (this._clientMode === ClientMode.OFFLINE) {
            return Promise.resolve(new JsonClientResponse(this.createDialogMessageModel(`GET action is not valid while offline: ${resourcePath}`), 400));
        }
        return this._fetchClient.getJson(baseUrl, resourcePath, queryParams);
    }
    getRecord(baseUrl, resourcePath, resourcePathElems, queryParams) {
        const pathFields = this.deconstructGetRecordPath(resourcePathElems);
        if (pathFields.dialogId === DialogProxy.BRIEFCASE_RECORD_DIALOG_ID) {
            let record = PersistenceTools.readRecordState(this._tenantId, this._userId, pathFields.dialogId);
            if (!record) {
                record = BriefcaseTemplate.BRIEFCASE_RECORD;
                PersistenceTools.writeRecordState(this._tenantId, this._userId, pathFields.dialogId, record);
            }
            return Promise.resolve(new JsonClientResponse(record, 200));
        }
        if (this._clientMode === ClientMode.OFFLINE) {
            const record = PersistenceTools.readRecordState(this._tenantId, this._userId, pathFields.dialogId);
            if (!record) {
                return Promise.resolve(new JsonClientResponse(this.createDialogMessageModel(`Offline record not found: ${pathFields.dialogId}`), 404));
            }
            else {
                return Promise.resolve(new JsonClientResponse(record, 200));
            }
        }
        const response = this._fetchClient.getJson(baseUrl, resourcePath, queryParams);
        return response.then(jcr => {
            if (jcr.statusCode === 200) {
                const record = jcr.value;
                const dialogId = pathFields.dialogId;
                const dialog = PersistenceTools.findRootDialogState(this._tenantId, this._userId, dialogId);
                // If we have persisted the dialog, we also need to persist the dialog records
                if (dialog) {
                    PersistenceTools.writeRecordState(this._tenantId, this._userId, dialogId, record);
                }
                return new JsonClientResponse(record, 200);
            }
            return response;
        });
    }
    getRedirection(baseUrl, resourcePath, resourcePathElems, queryParams) {
        const pathFields = this.deconstructGetRedirectionPath(resourcePathElems);
        if (pathFields.redirectionId === DialogProxy.BRIEFCASE_DIALOG_ID) {
            return Promise.resolve(new JsonClientResponse(BriefcaseTemplate.BRIEFCASE_REDIRECTION, 303));
        }
        if (this._clientMode === ClientMode.OFFLINE) {
            return Promise.resolve(new JsonClientResponse(this.createDialogMessageModel(`GET redirection is not valid while offline: ${resourcePath}`), 400));
        }
        return this._fetchClient.getJson(baseUrl, resourcePath, queryParams);
    }
    getSession(baseUrl, resourcePath, resourcePathElems, queryParams) {
        if (this._clientMode === ClientMode.OFFLINE) {
            return Promise.resolve(new JsonClientResponse(this.createDialogMessageModel(`GET session is not valid while offline: ${resourcePath}`), 400));
        }
        return this._fetchClient.getJson(baseUrl, resourcePath, queryParams);
    }
    getText(baseUrl, resourcePath) {
        if (this._clientMode === ClientMode.OFFLINE) {
            return Promise.resolve(new TextClientResponse(null, 400));
        }
        return this._fetchClient.getText(baseUrl, resourcePath);
    }
    isWorkPackagesDialog(dialog) {
        const referringObject = dialog.referringObject;
        return (referringObject &&
            referringObject.type === DialogProxy.REFERRING_WORKBENCH_MODEL_TYPE &&
            referringObject.actionId === DialogProxy.WORK_PACKAGES_ACTION_ID);
    }
    patchWorkPackagesDialog(dialog) {
        const workPackagesTableDialog = dialog.children[0];
        const propertyDefs = workPackagesTableDialog.recordDef.propertyDefs;
        propertyDefs.push({
            writeAllowed: false,
            propertyName: 'briefcase',
            canCauseSideEffects: false,
            upperCaseOnly: false,
            propertyType: 'boolean',
            type: 'hxgn.api.dialog.PropertyDef',
            writeEnabled: false
        });
        const columns = workPackagesTableDialog.view.columns;
        columns.push({
            propertyName: 'briefcase',
            heading: 'Briefcase',
            type: 'hxgn.api.dialog.Column'
        });
        return dialog;
    }
    patchWorkPackagesRecordSet(dialog, recordSet) {
        const existingRecordSet = this.readBriefcaseWorkpackageRecordSet();
        const existingRecords = existingRecordSet ? existingRecordSet.records : [];
        const records = recordSet.records;
        if (records) {
            for (const r of records) {
                let existingBriefcase = false;
                for (const existingRecord of existingRecords) {
                    if (r.id === existingRecord.id) {
                        existingBriefcase = true;
                        break;
                    }
                }
                const briefcaseField = {
                    name: 'briefcase',
                    annotations: [],
                    type: 'hxgn.api.dialog.Property',
                    value: existingBriefcase
                };
                r.properties.push(briefcaseField);
            }
        }
        return recordSet;
    }
    makeNullRedirectionId() {
        return `null_redirection__offline_${Date.now()}`;
    }
    postAddToBriefcaseMenuAction(baseUrl, resourcePath, resourcePathElems, jsonBody) {
        const pathFields = this.deconstructPostMenuActionPath(resourcePathElems);
        const redirectionId = this.makeNullRedirectionId;
        if (!jsonBody || !jsonBody.targets || jsonBody.targets.length === 0) {
            return Promise.resolve(new JsonClientResponse(this.createDialogMessageModel('Selection required'), 400));
        }
        const recordSet = PersistenceTools.readRecordSetState(this._tenantId, this._userId, pathFields.dialogId);
        if (!recordSet) {
            return Promise.resolve(new JsonClientResponse(this.createDialogMessageModel('Workpacakges not found'), 400));
        }
        const selectedRecords = [];
        for (const target of jsonBody.targets) {
            for (const record of recordSet.records) {
                if (record.id === target) {
                    PersistenceTools.updateRecordPropertyValue(record, 'briefcase', true);
                    selectedRecords.push(record);
                }
            }
        }
        PersistenceTools.writeRecordSetState(this._tenantId, this._userId, pathFields.dialogId, recordSet);
        const recordsAddedToBriefcase = [];
        for (const selected of selectedRecords) {
            const briefcaseProps = [];
            for (const p of selected.properties) {
                if (p.name === 'Disciplines') {
                    briefcaseProps.push({
                        name: 'disciplines',
                        format: null,
                        annotations: [],
                        type: 'hxgn.api.dialog.Property',
                        value: p.value
                    });
                }
                else if (p.name === 'Owning_Group') {
                }
                else if (p.name === 'Description') {
                    briefcaseProps.push({
                        name: 'description',
                        format: null,
                        annotations: [],
                        type: 'hxgn.api.dialog.Property',
                        value: p.value
                    });
                }
                else if (p.name === 'Config') {
                }
                else if (p.name === 'Name') {
                    briefcaseProps.push({
                        name: 'name',
                        annotations: [],
                        type: 'hxgn.api.dialog.Property',
                        value: p.value
                    });
                }
                else if (p.name === 'UID') {
                }
                else if (p.name === 'Creation_User') {
                }
                else if (p.name === 'Classification') {
                }
                else if (p.name === 'Organizations') {
                }
                else if (p.name === 'Creation_Date') {
                    briefcaseProps.push({
                        name: 'creation_date',
                        format: 'date',
                        annotations: [],
                        type: 'hxgn.api.dialog.Property',
                        value: p.value
                    });
                }
                else if (p.name === 'Last_Update_Date') {
                    briefcaseProps.push({
                        name: 'last_update_date',
                        format: 'date',
                        annotations: [],
                        type: 'hxgn.api.dialog.Property',
                        value: p.value
                    });
                }
                else if (p.name === 'Id') {
                    briefcaseProps.push({
                        name: 'workpackageid',
                        annotations: [],
                        type: 'hxgn.api.dialog.Property',
                        value: p.value
                    });
                }
                else if (p.name === 'Contract') {
                }
            }
            const nextAdd = {
                annotations: [],
                id: selected.id,
                type: 'hxgn.api.dialog.Record',
                properties: briefcaseProps
            };
            recordsAddedToBriefcase.push(nextAdd);
        }
        const briefcaseWorkpackagesRecordSet = this.readBriefcaseWorkpackageRecordSet();
        briefcaseWorkpackagesRecordSet.records = briefcaseWorkpackagesRecordSet.records.concat(recordsAddedToBriefcase);
        this.writeBriefcaseWorkpackageRecordSet(briefcaseWorkpackagesRecordSet);
        const actionParameters = {
            targets: [jsonBody.targets],
            type: 'hxgn.api.dialog.ActionParameters'
        };
        const nullRedirection = {
            tenantId: this._tenantId,
            referringObject: {
                dialogMode: 'LIST',
                dialogProperties: {
                    globalRefresh: 'true',
                    localRefresh: 'true',
                    dialogAlias: 'Workpackage_AddToBriefcase'
                },
                actionId: DialogProxy.ADD_TO_BRIEFCASE_ACTION_ID,
                type: 'hxgn.api.dialog.ReferringDialog',
                dialogId: pathFields.actionId
            },
            sessionId: pathFields.sessionId,
            id: redirectionId,
            type: 'hxgn.api.dialog.NullRedirection'
        };
        return Promise.resolve(new JsonClientResponse(nullRedirection, 303));
    }
    postJson(baseUrl, resourcePath, jsonBody) {
        const resourcePathElems = resourcePath.split('/');
        if (PersistenceTools.isPostSession(resourcePathElems)) {
            return this.postSession(baseUrl, resourcePath, resourcePathElems, jsonBody);
        }
        else if (PersistenceTools.isPostMenuAction(resourcePathElems)) {
            return this.postMenuAction(baseUrl, resourcePath, resourcePathElems, jsonBody);
        }
        else if (PersistenceTools.isPostWorkbenchAction(resourcePathElems)) {
            return this.postWorkbenchAction(baseUrl, resourcePath, resourcePathElems, jsonBody);
        }
        else if (PersistenceTools.isPostRecords(resourcePathElems)) {
            return this.postRecords(baseUrl, resourcePath, resourcePathElems, jsonBody);
        }
        if (this._clientMode === ClientMode.OFFLINE) {
            return Promise.resolve(new JsonClientResponse(this.createDialogMessageModel(`POST action is not valid while offline: ${resourcePath}`), 400));
        }
        return this._fetchClient.postJson(baseUrl, resourcePath, jsonBody);
    }
    postMenuAction(baseUrl, resourcePath, resourcePathElems, jsonBody) {
        const pathFields = this.deconstructPostMenuActionPath(resourcePathElems);
        if (pathFields.actionId === DialogProxy.ADD_TO_BRIEFCASE_ACTION_ID) {
            return this.postAddToBriefcaseMenuAction(baseUrl, resourcePath, resourcePathElems, jsonBody);
        }
        if (pathFields.actionId === DialogProxy.ENTER_OFFLINE_MODE_ACTION_ID ||
            pathFields.actionId === DialogProxy.EXIT_OFFLINE_MODE_ACTION_ID) {
            const redirectionId = this.makeNullRedirectionId();
            const nullRedirection = {
                tenantId: this._tenantId,
                referringObject: {
                    dialogMode: 'READ',
                    dialogProperties: {
                        globalRefresh: 'true',
                        localRefresh: 'true'
                    },
                    actionId: pathFields.actionId,
                    type: 'hxgn.api.dialog.ReferringDialog',
                    dialogId: pathFields.dialogId
                },
                sessionId: pathFields.sessionId,
                id: redirectionId,
                type: 'hxgn.api.dialog.NullRedirection'
            };
            return new Promise((resolve, reject) => {
                const online = pathFields.actionId === DialogProxy.EXIT_OFFLINE_MODE_ACTION_ID;
                const briefcaseRecord = this.readBriefcaseRecord();
                if (!briefcaseRecord) {
                    resolve(new JsonClientResponse(this.createDialogMessageModel('Briefcase not found'), 400));
                }
                PersistenceTools.updateRecordPropertyValue(briefcaseRecord, DialogProxy.ONLINE_PROPERTY_NAME, online);
                this.writeBriefcaseRecord(briefcaseRecord);
                this.changeClientMode(online ? ClientMode.ONLINE : ClientMode.OFFLINE);
                resolve(new JsonClientResponse(nullRedirection, 303));
            });
        }
        if (this._clientMode === ClientMode.OFFLINE) {
            const target = jsonBody.targets[0];
            let alias = null;
            if (pathFields.actionId == 'alias_Open') {
                const referringAlias = PersistenceTools.readDialogAliasState(this._tenantId, this._userId, pathFields.dialogId);
                alias = `Documents(${target})`;
                const navigationId = `${referringAlias}.${pathFields.actionId}.${alias}`;
                const navigation = PersistenceTools.readNavigationState(this._tenantId, this._userId, navigationId);
                if (navigation) {
                    const redirection = PersistenceTools.readRedirectionState(this._tenantId, this._userId, navigation.redirectionId);
                    if (redirection) {
                        return Promise.resolve(new JsonClientResponse(redirection, 303));
                    }
                    else {
                        return Promise.resolve(new JsonClientResponse(this.createDialogMessageModel(`Documents redirection not found at: ${pathFields.dialogId}`), 404));
                    }
                }
                else {
                    return Promise.resolve(new JsonClientResponse(this.createDialogMessageModel(`Documents navigation not found at: ${pathFields.dialogId}`), 404));
                }
            }
        }
        const response = this._fetchClient.postJson(baseUrl, resourcePath, jsonBody);
        return response.then(jcr => {
            if (jcr.statusCode === 303) {
                const redirection = jcr.value;
                if (redirection &&
                    redirection.referringObject &&
                    redirection.referringObject.type === DialogProxy.REFERRING_DIALOG_MODEL_TYPE) {
                    const target = jsonBody.targets[0];
                    const referringDialog = redirection.referringObject;
                    const rootDialog = PersistenceTools.findRootDialogState(this._tenantId, this._userId, referringDialog.dialogId);
                    if (rootDialog) {
                        const referringAlias = PersistenceTools.readDialogAliasState(this._tenantId, this._userId, referringDialog.dialogId);
                        let alias = null;
                        if (pathFields.actionId == 'alias_Open') {
                            alias = `Documents(${target})`;
                        }
                        else if (pathFields.actionId == 'alias_ShowTags') {
                            alias = `Tags(${target})`;
                        }
                        else if (pathFields.actionId == 'open') {
                            if (referringAlias.startsWith('Tags')) {
                                alias = `TagDetails(${target})`;
                            }
                            else if (referringAlias.startsWith('Documents')) {
                                alias = `DocumentDetails(${target})`;
                            }
                        }
                        const navigationId = alias
                            ? `${referringAlias}.${pathFields.actionId}.${alias}`
                            : `${referringAlias}.${pathFields.actionId}.not_available`;
                        let navigation = PersistenceTools.readNavigationState(this._tenantId, this._userId, navigationId);
                        if (navigation) {
                            PersistenceTools.deleteAllDialogState(this._tenantId, this._userId, navigation.redirectionId);
                        }
                        navigation = {
                            id: navigationId,
                            redirectionId: redirection.id,
                            redirectionAlias: alias
                        };
                        PersistenceTools.writeNavigationState(this._tenantId, this._userId, navigation);
                        PersistenceTools.writeRedirectionState(this._tenantId, this._userId, redirection);
                        if (alias) {
                            PersistenceTools.writeDialogAliasState(this._tenantId, this._userId, redirection.id, alias);
                        }
                    }
                }
            }
            return response;
        });
    }
    postMultipart(baseUrl, resourcePath, formData) {
        if (this._clientMode === ClientMode.OFFLINE) {
            return Promise.resolve(new VoidClientResponse(400));
        }
        return this._fetchClient.postMultipart(baseUrl, resourcePath, formData);
    }
    postRecords(baseUrl, resourcePath, resourcePathElems, jsonBody) {
        const pathFields = this.deconstructPostRecordsPath(resourcePathElems);
        if (pathFields.dialogId === DialogProxy.BRIEFCASE_WORKPACKAGES_DIALOG_ID) {
            let recordSet = PersistenceTools.readRecordSetState(this._tenantId, this._userId, pathFields.dialogId);
            if (!recordSet) {
                recordSet = BriefcaseTemplate.BRIEFCASE_WORKPACKAGES_RECORDSET;
                PersistenceTools.writeRecordSetState(this._tenantId, this._userId, pathFields.dialogId, recordSet);
            }
            return Promise.resolve(new JsonClientResponse(recordSet, 200));
        }
        if (this._clientMode === ClientMode.OFFLINE) {
            return new Promise((resolve, reject) => {
                const recordSet = PersistenceTools.readRecordSetState(this._tenantId, this._userId, pathFields.dialogId);
                if (!recordSet) {
                    resolve(new JsonClientResponse(this.createDialogMessageModel(`Offline record set not found ${pathFields.dialogId}`), 404));
                }
                else {
                    resolve(new JsonClientResponse(recordSet, 200));
                }
            });
        }
        const response = this._fetchClient.postJson(baseUrl, resourcePath, jsonBody);
        return response.then(jcr => {
            if (jcr.statusCode === 200) {
                let recordSet = jcr.value;
                const dialogId = pathFields.dialogId;
                const dialog = PersistenceTools.findRootDialogState(this._tenantId, this._userId, dialogId);
                // If we have persisted the dialog, we also need to persist the dialog records
                if (dialog) {
                    if (this.isWorkPackagesDialog(dialog)) {
                        recordSet = this.patchWorkPackagesRecordSet(dialog, recordSet);
                    }
                    PersistenceTools.writeRecordSetState(this._tenantId, this._userId, dialogId, recordSet);
                }
                return new JsonClientResponse(recordSet, 200);
            }
            return response;
        });
    }
    postSession(baseUrl, resourcePath, resourcePathElems, jsonBody) {
        this.changeTenantId(resourcePathElems[1]);
        this.changeUserId(jsonBody.userId);
        const briefcaseRecord = this.readBriefcaseRecord();
        if (briefcaseRecord) {
            const onlineProperty = PersistenceTools.findRecordProperty(briefcaseRecord, DialogProxy.ONLINE_PROPERTY_NAME);
            if (onlineProperty && !onlineProperty.value) {
                this.changeClientMode(ClientMode.OFFLINE);
            }
            else {
                this.changeClientMode(ClientMode.ONLINE);
            }
        }
        if (this._clientMode === ClientMode.OFFLINE) {
            return new Promise((resolve, reject) => {
                const session = PersistenceTools.readSessionState(this._tenantId, this._userId);
                if (!session) {
                    resolve(new JsonClientResponse(this.createOfflineSessionNotFoundModel(), 404));
                }
                else {
                    this.changeSessionId(session.id);
                    resolve(new JsonClientResponse(session, 200));
                }
            });
        }
        const response = this._fetchClient.postJson(baseUrl, resourcePath, jsonBody);
        return response.then(jcr => {
            if (jcr.statusCode === 200) {
                const session = jcr.value;
                this.changeSessionId(session.id);
                //                PersistenceTools.deleteAllState(this._tenantId, this._userId);
                PersistenceTools.writeSessionState(session);
            }
            return response;
        });
    }
    postWorkbenchAction(baseUrl, resourcePath, resourcePathElems, jsonBody) {
        const pathFields = this.deconstructPostWorkbenchActionPath(resourcePathElems);
        if (pathFields.actionId === DialogProxy.WORK_PACKAGES_ACTION_ID) {
            return this.postWorkbenchActionWorkPackages(baseUrl, resourcePath, resourcePathElems, jsonBody);
        }
        else if (pathFields.actionId === DialogProxy.BRIEFCASE_ACTION_ID) {
            return this.postWorkbenchActionBriefcase(baseUrl, resourcePath, resourcePathElems, jsonBody);
        }
        if (this._clientMode === ClientMode.OFFLINE) {
            return Promise.resolve(new JsonClientResponse(this.createDialogMessageModel(`Workbench action not valid while offline: ${pathFields.actionId}`), 400));
        }
        return this._fetchClient.postJson(baseUrl, resourcePath, jsonBody);
    }
    postWorkbenchActionBriefcase(baseUrl, resourcePath, resourcePathElems, jsonBody) {
        const pathFields = this.deconstructPostWorkbenchActionPath(resourcePathElems);
        const redirection = BriefcaseTemplate.BRIEFCASE_REDIRECTION;
        const navigationId = `${pathFields.actionId}Launcher`;
        const navigation = {
            id: navigationId,
            redirectionId: redirection.id
        };
        PersistenceTools.writeNavigationState(this._tenantId, this._userId, navigation);
        PersistenceTools.writeRedirectionState(this._tenantId, this._userId, redirection);
        PersistenceTools.writeDialogReferringAliasState(this._tenantId, this._userId, redirection.id, 'BriefcaseLauncher');
        PersistenceTools.writeDialogAliasState(this._tenantId, this._userId, redirection.id, 'Briefcase');
        return Promise.resolve(new JsonClientResponse(redirection, 303));
    }
    postWorkbenchActionWorkPackages(baseUrl, resourcePath, resourcePathElems, jsonBody) {
        const pathFields = this.deconstructPostWorkbenchActionPath(resourcePathElems);
        const navigationId = `${pathFields.actionId}Launcher`;
        if (this._clientMode === ClientMode.OFFLINE) {
            return new Promise((resolve, reject) => {
                const navigation = PersistenceTools.readNavigationState(this._tenantId, this._userId, navigationId);
                if (!navigation) {
                    resolve(new JsonClientResponse(this.createDialogMessageModel(`Navigation for offline ${pathFields.actionId} not found`), 404));
                }
                else {
                    const redirection = PersistenceTools.readRedirectionState(this._tenantId, this._userId, navigation.redirectionId);
                    if (!redirection) {
                        resolve(new JsonClientResponse(this.createDialogMessageModel(`Redirection for offline ${pathFields.actionId} not found`), 404));
                    }
                    else {
                        resolve(new JsonClientResponse(redirection, 303));
                    }
                }
            });
        }
        const response = this._fetchClient.postJson(baseUrl, resourcePath, jsonBody);
        return response.then(jcr => {
            if (jcr.statusCode === 303) {
                if (pathFields.actionId === DialogProxy.WORK_PACKAGES_ACTION_ID) {
                    const redirection = jcr.value;
                    PersistenceTools.deleteAllWorkbenchNavigation(this._tenantId, this._userId, navigationId);
                    const navigation = {
                        id: navigationId,
                        redirectionId: redirection.id
                    };
                    PersistenceTools.writeNavigationState(this._tenantId, this._userId, navigation);
                    PersistenceTools.writeRedirectionState(this._tenantId, this._userId, redirection);
                    PersistenceTools.writeDialogReferringAliasState(this._tenantId, this._userId, redirection.id, 'WorkPackagesLauncher');
                    PersistenceTools.writeDialogAliasState(this._tenantId, this._userId, redirection.id, 'WorkPackages');
                }
            }
            return response;
        });
    }
    putJson(baseUrl, resourcePath, jsonBody) {
        if (this._clientMode === ClientMode.OFFLINE) {
            return Promise.resolve(new JsonClientResponse(this.createDialogMessageModel(`PUT action is not valid while offline: ${resourcePath}`), 400));
        }
        return this._fetchClient.putJson(baseUrl, resourcePath, jsonBody);
    }
    /*
    * TODO: Refactor this into proxy-tools as a general method when discriminator and ids can be parameterized
    */
    readBriefcaseRecord() {
        // FIND DIALOG ID
        const briefcaseNavigationId = `${DialogProxy.BRIEFCASE_ACTION_ID}Launcher`;
        const briefcaseNavigation = PersistenceTools.readNavigationState(this._tenantId, this._userId, briefcaseNavigationId);
        if (briefcaseNavigation) {
            const redirectionId = briefcaseNavigation.redirectionId;
            const dialog = PersistenceTools.readDialogState(this._tenantId, this._userId, redirectionId);
            let briefcaseRecordDialogId = null;
            const dialogChildren = dialog.children;
            if (dialogChildren) {
                for (const child of dialogChildren) {
                    // TODO: We need a better discriminator technique for reading record(s) for a
                    // particular dialog.
                    if (child.businessClassName.endsWith('briefcase')) {
                        briefcaseRecordDialogId = child.id;
                        break;
                    }
                }
            }
            // READ RECORD
            if (briefcaseRecordDialogId) {
                return PersistenceTools.readRecordState(this._tenantId, this._userId, briefcaseRecordDialogId);
            }
        }
        return null;
    }
    readBriefcaseWorkpackageRecordSet() {
        return PersistenceTools.readRecordSetState(this._tenantId, this._userId, DialogProxy.BRIEFCASE_WORKPACKAGES_DIALOG_ID);
    }
    setClientMode(clientMode) {
        this.changeClientMode(clientMode);
    }
    writeBriefcaseWorkpackageRecordSet(recordSet) {
        return PersistenceTools.writeRecordSetState(this._tenantId, this._userId, DialogProxy.BRIEFCASE_WORKPACKAGES_DIALOG_ID, recordSet);
    }
    /*
    * TODO: Refactor this into proxy-tools as a general method when discriminator and ids can be parameterized
    */
    writeBriefcaseRecord(briefcaseRecord) {
        // FIND DIALOG ID
        const briefcaseNavigationId = `${DialogProxy.BRIEFCASE_ACTION_ID}Launcher`;
        const briefcaseNavigation = PersistenceTools.readNavigationState(this._tenantId, this._userId, briefcaseNavigationId);
        if (briefcaseNavigation) {
            const redirectionId = briefcaseNavigation.redirectionId;
            const dialog = PersistenceTools.readDialogState(this._tenantId, this._userId, redirectionId);
            let briefcaseRecordDialogId = null;
            const dialogChildren = dialog.children;
            if (dialogChildren) {
                for (const child of dialogChildren) {
                    // TODO: We need a better discriminator technique for reading record(s) for a
                    // particular dialog.
                    if (child.businessClassName.endsWith('briefcase')) {
                        briefcaseRecordDialogId = child.id;
                        break;
                    }
                }
            }
            // WRITE RECORD
            if (briefcaseRecordDialogId) {
                PersistenceTools.writeRecordState(this._tenantId, this._userId, briefcaseRecordDialogId, briefcaseRecord);
            }
        }
    }
}
DialogProxy.ADD_TO_BRIEFCASE_ACTION_ID = 'alias_AddToBriefcase';
DialogProxy.BRIEFCASE_ACTION_ID = 'Briefcase';
DialogProxy.BRIEFCASE_DIALOG_ID = 'a1';
DialogProxy.BRIEFCASE_RECORD_DIALOG_ID = 'a11';
DialogProxy.BRIEFCASE_WORKPACKAGES_DIALOG_ID = 'a12';
DialogProxy.ENTER_OFFLINE_MODE_ACTION_ID = 'alias_EnterOfflineMode';
DialogProxy.EXIT_OFFLINE_MODE_ACTION_ID = 'alias_ExitOfflineMode';
DialogProxy.OPEN_ACTION_ID = 'open';
DialogProxy.OPEN_DOCUMENT_ACTION_ID = 'dynamic_AAABACcaAAAAACqd';
DialogProxy.ONLINE_PROPERTY_NAME = 'online';
DialogProxy.REFERRING_WORKBENCH_MODEL_TYPE = 'hxgn.api.dialog.ReferringWorkbench';
DialogProxy.REFERRING_DIALOG_MODEL_TYPE = 'hxgn.api.dialog.ReferringDialog';
DialogProxy.SDA_WORKBENCH_ID = 'AAABACffAAAAACe2';
DialogProxy.WORK_PACKAGES_ACTION_ID = 'WorkPackages';
